%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Computing the completion of a given logic program               %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v1.0: 07/11/2003 by Ulle Endriss (ue@doc.ic.ac.uk)              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This module requires the predicate construct_equalities/3, which 
%%% is provided by ciff-aux.pl. The main predicate of the module is
%%% close_program/4.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load Sicstus Prolog library for lists.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library(lists)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Declare iff-connective as an operator.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- op( 500, yfx, iff).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% close_program( +CorePreds, +NonPreds, +Listing, -Definitions).
%%% Given the Listing of a logic program as a list, return the list
%%% of iff-definitions that constitute the completion of that program.
%%% In addition, for every predicate given in the list CorePreds,
%%% also include an iff-definition in the answer, even if the predicate 
%%% does not feature in the listing. Also, do not include a definition
%%% for any predicate in the list NonPreds. The elements of CorePreds
%%% and NonPreds are of the for p/2 etc.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close_program( CorePreds, NonPreds, Listing, Defs) :-
  get_preds( CorePreds, NonPreds, Listing, AdditionalPreds),
  append( CorePreds, AdditionalPreds, Preds),
  close_preds( Preds, Listing, Defs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% close_pred( +Pred/Arity, +Listing, -Definition).
%%% Return the iff-definition for the predicate specified as the first
%%% argument for the logic program provided as a list in Listing.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close_pred( Pred/Arity, [], PredVars iff []) :- 
  length( Vars, Arity),
  PredVars =.. [Pred|Vars].

close_pred( Pred/Arity, [Fact|Clauses], PredVars iff [Eqs|Def]) :-
  Fact =.. [Pred|Args],
  \+ ( Pred = (:-)), 
  length( Args, Arity), !,  
  close_pred( Pred/Arity, Clauses, PredVars iff Def),
  PredVars =.. [Pred|Vars],
  construct_equalities( Vars, Args, Eqs).

close_pred( Pred/Arity, [Rule|Clauses], PredVars iff [List|Def]) :-
  Rule = (Head :- Body),
  Head =.. [Pred|Args],
  length( Args, Arity), !,  
  close_pred( Pred/Arity, Clauses, PredVars iff Def),
  PredVars =.. [Pred|Vars],
  construct_equalities( Vars, Args, Eqs),
  nested_to_list( Body, BodyList),
  append( Eqs, BodyList, List).

close_pred( Pred/Arity, [_|Clauses], Closure) :-
  close_pred( Pred/Arity, Clauses, Closure). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% close_preds( +Preds, +Listing, -Definitions).
%%% Use close_pred/3 with a list of predicates to get a list of 
%%% iff-definitions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close_preds( [], _, []).

close_preds( [Pred|Preds], Listing, [Def|Defs]) :-
  close_pred( Pred, Listing, Def),
  close_preds( Preds, Listing, Defs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% get_preds( +ControlList, +RemoveList, +Listing, -Predicates).
%%% Get the list of all predicates occurring in Listing (a logic 
%%% program provided as a list of clauses), including at least the
%%% predicates mentioned in the list ControlList, and omitting those
%%% in RemoveList. Predicates in ControlList, RemoveList, and 
%%% Predicates are given in the form p/2 etc.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_preds( _, _, [], []).

get_preds( ControlList, RemoveList, [Rule|Clauses], Preds) :-
  Rule =.. [:-,Head|_], !,
  get_preds( ControlList, RemoveList, [Head|Clauses], Preds).

get_preds( ControlList, RemoveList, [Fact|Clauses], Preds) :-
  Fact =.. [Pred|Args],
  length( Args, Arity),
  (member( Pred/Arity, ControlList) ; member(Pred/Arity, RemoveList)), !,
  get_preds( ControlList, RemoveList, Clauses, Preds).

get_preds( ControlList, RemoveList, [Fact|Clauses], [Pred/Arity|Preds]) :-
  Fact =.. [Pred|Args],
  length( Args, Arity),
  get_preds( [Pred/Arity|ControlList], RemoveList, Clauses, Preds).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% nested_to_list( +Nested, -List).
%%% An auxiliary predicate that turn a nesting of pairs (a,b,c,...)
%%% as it may occur in the body of a clause into a list [a,b,c,...].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nested_to_list( (Elem,Nested), [Elem|List]) :- !,
  nested_to_list( Nested, List).

nested_to_list( Elem, [Elem]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
